// PathTruder.cpp 
//

// Path Extruder:
// Extrudes a shape along a path. No generation of conditional lines is done, Edger2 is your friend!
// Path and shape are provided as LDraw files containing lines.
// Both path files and both shape files must have the same number of lines.
// usage: pathtruder -p1 LDrawPath1 -p2 LDrawPath2 -s1 LDrawShape1 [-s2 LDrawShape2] [options] LdrawSurfaceFileOut

// Philippe "Philo" Hurbain - 2008 - www.philohome.com

#include "stdafx.h"

#include <math.h>
#include <string.h>

#define EPSILON 0.000001
#define VERTMERGE 0.001
#define PI 3.14159265358979323846

/* Null vector */

double null[3] = {0,0,0};

/* some macros */

#define CROSS(dest,v1,v2)                      \
              dest[0]=v1[1]*v2[2]-v1[2]*v2[1]; \
              dest[1]=v1[2]*v2[0]-v1[0]*v2[2]; \
              dest[2]=v1[0]*v2[1]-v1[1]*v2[0];

#define DOT(v1,v2) (v1[0]*v2[0]+v1[1]*v2[1]+v1[2]*v2[2])

#define SUB(dest,v1,v2) dest[0]=v1[0]-v2[0]; dest[1]=v1[1]-v2[1]; dest[2]=v1[2]-v2[2]; 

#define ADD(dest,v1,v2) dest[0]=v1[0]+v2[0]; dest[1]=v1[1]+v2[1]; dest[2]=v1[2]+v2[2];

#define MULT(dest,v,factor) dest[0]=factor*v[0]; dest[1]=factor*v[1]; dest[2]=factor*v[2];

#define SET(dest,src) dest[0]=src[0]; dest[1]=src[1]; dest[2]=src[2]; 

#define MANHATTAN(v1, v2) (fabs(v1[0]-v2[0]) + fabs(v1[1]-v2[1]) + fabs(v1[2]-v2[2]))

#define DIST(v1, v2) (sqrt((v1[0]-v2[0])*(v1[0]-v2[0]) + (v1[1]-v2[1])*(v1[1]-v2[1]) + (v1[2]-v2[2])*(v1[2]-v2[2])))


// Tri_Angle computes the angle (in degrees) between the planes of a quad.
// They are assumed to be non-degenerated
double Tri_Angle(double U0[3],double U1[3],double U2[3], double U3[3])
{
	double Unorm[3], Vnorm[3];
	double Temp[3];
	double U10[3], U20[3];
	double V10[3], V20[3];
	double len;
	SUB(U10, U1, U0);
	SUB(U20, U2, U0);
	SUB(V10, U2, U0);
	SUB(V20, U3, U0);
	CROSS(Temp, U10, U20);
	len = DIST(Temp, null);
	MULT(Unorm, Temp, 1/len);
	CROSS(Temp, V10, V20);
	len = DIST(Temp, null);
	MULT(Vnorm, Temp, 1/len);
	CROSS(Temp, Unorm, Vnorm);
	double dist;
	dist = DIST(Temp, null);
	if(dist>0.9999999999) return 90;
	return(180/PI*asin(dist));
}

// Calculate local basis, based on the direction of the i-th vector between both paths, 
// and the average of the planes defined by the paths before and after this vector
// Returns angle between these planes.


double PathLocalBasis (int n, int i,  double xv[],double yv[],double zv[], double path1[][2][3], double path2[][2][3])
{	
	double a, scale;
	double temp1[3], temp2[3], temp3[3], temp4[3];

	// Calculate local coordinate basis
	scale = DIST(path2[i][0], path1[i][0]);

	if(scale < EPSILON)
	{
		// size is 0... any non-degenerated base will do!
		SET (yv, null);
		yv[0]=1;
	}
	else
	{	
		SUB(yv, path1[i][0], path2[i][0]);
	}

	// Average Path Normal	
	SUB(temp1, path1[i][1], path1[i][0]);
	SUB(temp2, path2[i][1], path1[i][0]);
	CROSS(xv, temp2, temp1);
	a=DIST(xv, null);
	if(a>EPSILON) { MULT(xv, xv, (1/a)); } else { SET(xv, null);}
	SUB(temp1, path2[i][1], path2[i][0]);
	SUB(temp2, path1[i][0], path2[i][0]);
	CROSS(temp3, temp1, temp2);
	a=DIST(temp3, null);
	if(a>EPSILON) {MULT(temp3, temp3, (1/a));} else {SET(temp3, null);}
	ADD(xv, xv, temp3);
	a=DIST(xv, null);
	if(a>EPSILON) { MULT(xv, xv, (1/a)); } else { SET(xv, null);} 

	SUB(temp1, path1[n][1], path1[n][0]);
	SUB(temp2, path2[n][1], path1[n][0]);
	CROSS(temp4, temp2, temp1);
	a=DIST(temp4, null);
	if(a>EPSILON) {MULT(temp4, temp4, (1/a));} else {SET(temp4, null);}
	SUB(temp1, path2[n][1], path2[n][0]);
	SUB(temp2, path1[n][0], path2[n][0]);
	CROSS(temp3, temp1, temp2);
	a=DIST(temp3, null);
	if(a>EPSILON) {MULT(temp3, temp3, (1/a));} else {SET(temp3, null);}
	ADD(temp4, temp4, temp3);
	a=DIST(temp4, null);
	if(a>EPSILON) {MULT(temp4, temp4, (1/a));} else {SET(temp4, null);} 

	// Average previous and current path normals
	ADD(xv, xv, temp4);
	a=DIST(xv, null);
	if(a>EPSILON) { MULT(xv, xv, (1/a)); } else { SET(xv, null);} 

	// calculate angle
	a = 360 /PI * acos(DOT(xv,temp4));
	
	CROSS(zv,xv,yv);
	MULT(xv, xv, scale);
	if(scale < EPSILON) 
	{
		SET (yv, null);
		SET (zv, null);
	}
	return a;
}


// Calculate scaled sigmoid function between 0 and 1. 
// 1/(1+exp(-b*(x-m))) Scaled so that sigmoid(0)=0, sigmoid(1)=1
// b is growth rate, m is max growth rate point
// if b=1. returns a true x (linear relationship)
double sigmoid(double x,double b,double m)
{
	double s0, s1, y;
	if(b==1.) return x;
	s0=1/(1+exp(b*m));
	s1=1/(1+exp(-b*(1-m)));
	y=1/(1+exp(-b*(x-m)));
	y=(y-s0)/(s1-s0);
	return y;
}
	


void usage(void)
{
	printf ("Usage:\npathtruder -p1 Path1 -p2 Path2 -s1 Shape1 [-s2 Shape2] [options] SurfaceFileOut\n");
	printf ("  -p1 <LDraw path file 1> defines the main path\n"); 
	printf ("  -p2 <LDraw path file 2> defines the secondary path\n");
	printf ("  -s1 <LDraw shape file 1> defines the shape to be swept along the path\n");
	printf ("  -s2 <LDraw shape file 2> defines the ending shape\n");
	printf ("      If -s1 and -s2 are defined, the shape is progressively morphed\n");
	printf ("      between these shapes along the path.\n");
	printf ("      This transition can be controlled with -t options:\n");
	printf ("   -tn <number> sets the number of transitions from shape1 to shape2\n");
    printf ("       and vice versa. 1 (default) sweeps from s1 to s2,\n"); 
	printf ("       2 from s1 to s2 then back to s1, and so on\n");
	printf ("   -ts <slope> controls transition curve, from 1 (linear)\n"); 
	printf ("       to 10 (s-shape, default) to 100 (step)\n");
	printf ("   -tp <position> controls transition curve centering.\n"); 
	printf ("   	0 <= position <= 1. Default is 0.5, centered\n");
	printf ("   -e Use path start and end segments to define orientation of\n");
	printf ("      corresponding extruded section\n");
	printf ("   -i Invert curvature of shape\n");
	printf ("   -c <crease angle> If path has a sharper angle, a line is created\n");
	printf ("      at the junction. Default is 180 (no line).\n");
	printf ("      Use a negative value to force line creation\n");
	printf ("   -a When specified, shape is elongated to compensate for flatening\n");
	printf ("      at sharp path angles. Default is no compensation\n");
	printf ("   -r <rotation angle> specifies rotation of shape direction vector around\n");
	printf ("      path1 as it sweeps along the path. Default is 0 degrees\n");
	printf ("   -l <length> define maximum path segment length. Longer segments are split\n");
	printf("Press <Enter> to quit");
	getchar();
	
/*	printf ("   -ra <rotation angle> specifies rotation of shape direction vector around\n");
	printf ("       path1 as it sweeps along the path. Default is 0\n");
	printf ("   -rs <slope> define angle change curve (default is 1, linear)\n");
	printf ("   -rp <position> controls centering of angle change curve\n");
	printf ("       (default is 0.5, centered)\n");
*/	
}

#define MAX_LINE 1000
#define SMALL 0.1
#define SMALLANGLE .95


int main(int argc, char* argv[])
{
	double Path1 [5*MAX_LINE][2][3];
	double Path2[5*MAX_LINE][2][3];
	double Path1a[MAX_LINE][2][3];
	double Path2a[MAX_LINE][2][3];

	double Shape1[MAX_LINE][2][3];
	double Shape2[MAX_LINE][2][3];
	double CurShape[MAX_LINE][2][3];
	double NxtShape[MAX_LINE][2][3];
	double Shape1Vect[3], Shape2Vect[3];

	double temp1[3], temp2[3], temp3[3];
	double XVect[3], YVect[3], ZVect[3];

	double Angle, ca, sa;
	double ratio;
	double SortBuf[MAX_LINE][2][3];
	int next [MAX_LINE][2][2];

	FILE *LdrawFileOut;

	int Path1Flag=false;
	int Path2Flag=false;
	int Shape1Flag=false;
	int Shape2Flag=false;

	int Path1Len=0;
	int Path2Len=0;
	int Shape1Len=0;
	int Shape2Len=0;

	FILE *LdrawFileIn;

	char buf[1024];
	int type, color;

	int circular;
	double maxlength = 100000;
	double d, dmax;
	double len;
	int InLineIdx;
	int NumPath;
	bool invert=false;
	int i,j,k,l;
	int n;

	int transitions=1;
	double slope=10;
	double position=0.5;
	double crease = 180;
	int compensate = false;
	int endings=false;
	double rotation = 0;
	double rotslope=1;
	double rotpos=0.5;
	double rotangle;

	printf("\nLdraw PathTruder v1.2 - by Philo");
	printf("\n--------------------------------\n");

	if(argc < 8) 		
	{
		usage();
		return 1;
	}
	for(n=1; n<argc-1; n++)
	{
		if(argv[n][0] != '-') 
		{
			usage();
			return 1;
		}
		switch (argv[n][1])
		{

		case 'l':
			n++;
			sscanf(argv[n], "%lf", &maxlength);
			if (maxlength<=0)
			{
				printf("Length (-l %.7lg) must be > 0\n", maxlength);
				printf("Press <Enter> to quit");
				getchar();
				return 3;
			}
			printf("l = %.7lg\n", maxlength);
			break;

		case 'r':
			n++;
			sscanf(argv[n], "%lf", &rotation);
			printf("Spinning angle = %.7lg\n", rotation);
/*			switch (argv[n][2])
			{
			case 'a':
				n++;
				sscanf(argv[n], "%lf", &rotation);
				printf("Spinning angle = %.7lg\n", rotation);
				break;
			case 's':
				n++;
				sscanf(argv[n], "%lf", &rotslope);
				if (rotslope<1)
				{
					printf("Spinning slope (-rs %.7lf) must be >= 1\n", rotslope);
					printf("Press <Enter> to quit");
					getchar();
					return 3;
				}
				printf("Spinning slope = %.7lf\n", rotslope);
				break;
			case 'p':
				n++;
				sscanf(argv[n], "%lf", &rotpos);
				if ((rotpos>1) || (rotpos<0))
				{
					printf("Spinning center position (-rp %.7lf) must be in [0..1]\n", rotpos);
					printf("Press <Enter> to quit");
					getchar();
					return 3;
				}
				printf("Spinning center position = %.7lf\n", position);
				break;
			default:
				usage();
				return 2;
				break;
			}
*/
			break;

		case 'a':
			printf("compensate sharp angles\n");
			compensate=true;
			break;
		case 'i':
			printf("invert shape curvature\n");
			invert=true;
			break;
		case 'c':
			n++;
			sscanf(argv[n], "%lf", &crease);
			printf("crease angle = %.7lg\n", crease);
			break;
		case 'e':
			printf("Ends of path used for orientation\n");
			endings=true;
			break;
		case 't':
			switch (argv[n][2])
			{
			case 'n':
				n++;
				sscanf(argv[n], "%d", &transitions);
				if (transitions<=0)
				{
					printf("Number of shape1 to shape2 transitions (-tn %d) must be > 0\n", transitions);
					printf("Press <Enter> to quit");
					getchar();
					return 3;
				}
				printf("transitions = %d\n", transitions);
				break;
			case 's':
				n++;
				sscanf(argv[n], "%lf", &slope);
				if (slope<1)
				{
					printf("Transitions slope (-ts %.7lf) must be >= 1\n", slope);
					printf("Press <Enter> to quit");
					getchar();
					return 3;
				}
				printf("Transition slope = %.7lf\n", slope);
				break;
			case 'p':
				n++;
				sscanf(argv[n], "%lf", &position);
				if ((position>1) || (position<0))
				{
					printf("Transitions position (-tp %.7lf) must be in [0..1]\n", position);
					printf("Press <Enter> to quit");
					getchar();
					return 3;
				}
				printf("Transition position = %.7lf\n", position);
				break;
			default:
				usage();
				return 2;
				break;
			}
			break;
		case 'p':
			switch (argv[n][2])
			{
			case '1':
				Path1Flag=true;

				n++;
				LdrawFileIn = fopen (argv[n], "rt");
				if(LdrawFileIn == NULL)
				{
					printf("Can't open file\n%s\n", argv[n]);
					printf("Press <Enter> to quit");
					getchar();
					return 2;
				}
				printf("Read path file 1\n");


				while (fgets(buf, 1024, LdrawFileIn))
				{
					type = -1;
					sscanf(buf, "%d", &type);
					if(type==2)
					{
						sscanf (buf, "%d %d %lf %lf %lf %lf %lf %lf", &type, &color, 
							&SortBuf[Path1Len][0][0], &SortBuf[Path1Len][0][1], &SortBuf[Path1Len][0][2], 
							&SortBuf[Path1Len][1][0], &SortBuf[Path1Len][1][1], &SortBuf[Path1Len][1][2]);
						next[Path1Len][0][0] = next[Path1Len][1][0] = -1;
						if(MANHATTAN(SortBuf[Path1Len][0], SortBuf[Path1Len][1]) < EPSILON )
						{
							printf("Null length element in path file 1\n");
							printf("Press <Enter> to quit");
							getchar();
							return 3;
						}

						Path1Len++;
					}
					if(Path1Len>MAX_LINE)
					{
						printf("Too many lines in path file 1\n");
						printf("Press <Enter> to quit");
						getchar();
						return 3;
					}
				}
				if(Path1Len==0)
				{
					printf("No line in path file 1!\n");
					printf("Press <Enter> to quit");
					getchar();
					return 3;
				}
				printf("Sort path file 1\n");
				circular=true;
				for (i=0; i<Path1Len; i++)
				{
					for (j=0; j<2; j++)
					{
						if(next[i][j][0] != -1) break;
						dmax=100000;
						for (k=0; k<Path1Len; k++)
						{
							if(k != i)
							{
								for(l=0; l<2; l++)
								{
									d=MANHATTAN(SortBuf[i][j], SortBuf[k][l]);
									if(d<dmax)
									{
										dmax=d;
										next[i][j][0]=k;
										next[i][j][1]=l;
									}
									if(d==0) break;
								}
								if(d==0) break;
							}
						}
						if(dmax>SMALL) 
						{
							next[i][j][0]=-1;
							circular=false;
						}
					}
				}
				if(circular) 
				{
					next[next[0][0][0]][next[0][0][1]][0]=-1;
					next[0][0][0]=-1;
				}
				InLineIdx=0;
				NumPath=0;
				for (i=0; i<Path1Len; i++)
				{
					for (j=0; j<2; j++)
					{		
						int a,b,c,d;
						if(next[i][j][0] == -1)
						{
							NumPath++;
							a=i; b=j;
							do
							{
								SET(Path1a[InLineIdx][0], SortBuf[a][b]);
								SET(Path1a[InLineIdx][1], SortBuf[a][1-b]);
								InLineIdx++;

								d=next[a][1-b][1];
								c=next[a][1-b][0];
								next[a][1-b][0] = -2;
								next[a][b][0] = -2;
								b=d;
								a=c;
							}
							while((a!=-1));
						}
					}
				}

				Path1Len=InLineIdx;

				if(NumPath > 1)
					printf("%d distinct paths found in Path file 1. Unexpected results may happen!\n",NumPath);

				printf("%d lines in path file 1\n", Path1Len);
				fclose (LdrawFileIn);
				break;

			case '2':
				Path2Flag=true;

				n++;
				LdrawFileIn = fopen (argv[n], "rt");
				if(LdrawFileIn == NULL)
				{
					printf("Can't open file\n%s\n", argv[n]);
					printf("Press <Enter> to quit");
					getchar();
					return 2;
				}
				printf("Read path file 2\n");

				while (fgets(buf, 1024, LdrawFileIn))
				{
					type = -1;
					sscanf(buf, "%d", &type);
					if(type==2)
					{
						sscanf (buf, "%d %d %lf %lf %lf %lf %lf %lf", &type, &color, 
							&SortBuf[Path2Len][0][0], &SortBuf[Path2Len][0][1], &SortBuf[Path2Len][0][2], 
							&SortBuf[Path2Len][1][0], &SortBuf[Path2Len][1][1], &SortBuf[Path2Len][1][2]);
						next[Path2Len][0][0] = next[Path2Len][1][0] = -1;
						if(MANHATTAN(SortBuf[Path2Len][0], SortBuf[Path2Len][1]) < EPSILON )
						{
							printf("Null length element in path file 2\n");
							printf("Press <Enter> to quit");
							getchar();
							return 3;
						}

						Path2Len++;
					}
					if(Path2Len>MAX_LINE)
					{
						printf("Too many lines in path file 2\n");
						printf("Press <Enter> to quit");
						getchar();
						return 3;
					}
				}
				if(Path2Len==0)
				{
					printf("No line in path file 2!\n");
					printf("Press <Enter> to quit");
					getchar();
					return 3;
				}
				printf("Sort path file 2\n");
				circular=true;
				for (i=0; i<Path2Len; i++)
				{
					for (j=0; j<2; j++)
					{
						if(next[i][j][0] != -1) break;
						dmax=100000;
						for (k=0; k<Path2Len; k++)
						{
							if(k != i)
							{
								for(l=0; l<2; l++)
								{
									d=MANHATTAN(SortBuf[i][j], SortBuf[k][l]);
									if(d<dmax)
									{
										dmax=d;
										next[i][j][0]=k;
										next[i][j][1]=l;
									}
									if(d==0) break;
								}
								if(d==0) break;
							}
						}
						if(dmax>SMALL) 
						{
							next[i][j][0]=-1;
							circular=false;
						}
					}
				}
				if(circular) 
				{
					next[next[0][0][0]][next[0][0][1]][0]=-1;
					next[0][0][0]=-1;
				}
				InLineIdx=0;
				NumPath=0;
				for (i=0; i<Path2Len; i++)
				{
					for (j=0; j<2; j++)
					{		
						int a,b,c,d;
						if(next[i][j][0] == -1)
						{
							NumPath++;
							a=i; b=j;
							do
							{
								SET(Path2a[InLineIdx][0], SortBuf[a][b]);
								SET(Path2a[InLineIdx][1], SortBuf[a][1-b]);
								InLineIdx++;

								d=next[a][1-b][1];
								c=next[a][1-b][0];
								next[a][1-b][0] = -2;
								next[a][b][0] = -2;
								b=d;
								a=c;
							}
							while((a!=-1));
						}
					}
				}

				Path2Len=InLineIdx;

				if(NumPath > 1)
					printf("%d distinct paths found in Path file 2. Unexpected results may happen!\n",NumPath);
		
				printf("%d lines in path file 2\n", Path2Len);
				fclose (LdrawFileIn);
				break;
			default:
				usage();
				return 1;
				break;
			}
			break;
		case 's':
			switch (argv[n][2])
			{
			case '1':
				Shape1Flag=true;

				n++;
				LdrawFileIn = fopen (argv[n], "rt");
				if(LdrawFileIn == NULL)
				{
					printf("Can't open file\n%s\n", argv[n]);
					printf("Press <Enter> to quit");
					getchar();
					return 2;
				}
				printf("Read shape file 1\n");

				while (fgets(buf, 1024, LdrawFileIn))
				{
					type = -1;
					sscanf(buf, "%d", &type);
					if(type==2)
					{
						sscanf (buf, "%d %d %lf %lf %lf %lf %lf %lf", &type, &color, 
							&Shape1[Shape1Len][0][0], &Shape1[Shape1Len][0][1], &Shape1[Shape1Len][0][2], 
							&Shape1[Shape1Len][1][0], &Shape1[Shape1Len][1][1], &Shape1[Shape1Len][1][2]);
						Shape1Len++;
					}
					if(Shape1Len>MAX_LINE)
					{
						printf("Too many lines in shape file 1\n");
						printf("Press <Enter> to quit");
						getchar();
						return 3;
					}
				}
				printf("%d lines in shape file 1\n", Shape1Len);
				fclose (LdrawFileIn);
				break;
			case '2':
				Shape2Flag=true;

				n++;
				LdrawFileIn = fopen (argv[n], "rt");
				if(LdrawFileIn == NULL)
				{
					printf("Can't open file\n%s\n", argv[n]);
					printf("Press <Enter> to quit");
					getchar();
					return 2;
				}
				printf("Read shape file 2\n");

				while (fgets(buf, 1024, LdrawFileIn))
				{
					type = -1;
					sscanf(buf, "%d", &type);
					if(type==2)
					{
						sscanf (buf, "%d %d %lf %lf %lf %lf %lf %lf", &type, &color, 
							&Shape2[Shape2Len][0][0], &Shape2[Shape2Len][0][1], &Shape2[Shape2Len][0][2], 
							&Shape2[Shape2Len][1][0], &Shape2[Shape2Len][1][1], &Shape2[Shape2Len][1][2]);
						Shape2Len++;
					}
					if(Shape2Len>MAX_LINE)
					{
						printf("Too many lines in shape file 2\n");
						printf("Press <Enter> to quit");
						getchar();
						return 3;
					}
				}
				printf("%d lines in shape file 2\n", Shape2Len);
				fclose (LdrawFileIn);
				break;
			default:
				usage();
				return 1;
				break;
			}
			break;
		default:
			usage();
			return 1;
		}
	}

	if(n==argc)
	{
		printf("No output file specified!\n");
		printf("Press <Enter> to quit");
		getchar();
		return 2;
	}

	if(!Shape1Flag && !Shape2Flag)
	{
		printf("No shape file provided!\n");
		printf("Press <Enter> to quit");
		getchar();
		return 2;
	}
	if(!Shape1Flag)
	{
		printf("Shape 1 file not provided, copying shape 2 to shape 1\n");
		for(i=0; i<Shape2Len; i++)
		{
			SET(Shape1[i][0], Shape2[i][0]);
			SET(Shape1[i][1], Shape2[i][1]);
		}
		Shape1Len = Shape2Len;
	}
	if(!Shape2Flag)
	{
		printf("Shape 2 file not provided, copying shape 1 to shape 2\n");
		for(i=0; i<Shape1Len; i++)
		{
			SET(Shape2[i][0], Shape1[i][0]);
			SET(Shape2[i][1], Shape1[i][1]);
		}
		Shape2Len = Shape1Len;
	}

	if(!Path1Flag || !Path2Flag)
	{
		printf("Two path files are needed!\n");
		printf("Press <Enter> to quit");
		getchar();
		return 2;
	}

	if(Path1Len != Path2Len)
	{
		printf("The two path files do not have the same number of elements!\n");
		printf("Press <Enter> to quit");
		getchar();
		return 2;
	}

	if(endings && (Path1Len<3) && !circular)
	{
		printf("Path files must have at least 3 elements to use -e option!\n");
		printf("Press <Enter> to quit");
		getchar();
		return 2;
	}


	if(Shape1Len != Shape2Len)
	{
		printf("The two shape files do not have the same number of elements!\n");
		printf("Press <Enter> to quit");
		getchar();
		return 2;
	}


	LdrawFileOut = fopen (argv[argc-1], "wt");
	if(LdrawFileOut == NULL)
	{
		printf("Can't create file\n%s\n", argv[argc-1]);
		printf("Press <Enter> to quit");
		getchar();
		return 2;
	}

	printf("Split long lines\n");
	InLineIdx = 0;
	for(i=0; i<Path1Len; i++)
	{
		double p1[3], p2[3], q1[3], q2[3], delta1[3], delta2[3], temp[3];
		int nsplit1, nsplit2;

		SET(p1, Path1a[i][0]);
		SET(p2, Path1a[i][1]);

		SET(q1, Path2a[i][0]);
		SET(q2, Path2a[i][1]);

		nsplit1=(DIST(p1,p2)/maxlength)+1;
		nsplit2=(DIST(q1,q2)/maxlength)+1;

		// don't split endings segments
		if(endings)
		{
			if(i==0 || i==Path1Len-1) nsplit1 = nsplit2 = 1;
		}

		nsplit1 = (nsplit1 > nsplit2) ? nsplit1 : nsplit2;

		SUB(delta1,p2,p1);
		MULT(delta1, delta1, (1.0/nsplit1));
		SUB(delta2,q2,q1);
		MULT(delta2, delta2, (1.0/nsplit1));
		for (k=0; k<nsplit1; k++)
		{
			MULT (temp, delta1, k);
			ADD (Path1[InLineIdx][0],p1,temp);
			ADD (Path1[InLineIdx][1],Path1[InLineIdx][0],delta1);
			MULT (temp, delta2, k);
			ADD (Path2[InLineIdx][0],q1,temp);
			ADD (Path2[InLineIdx][1],Path2[InLineIdx][0],delta2);

			InLineIdx++;
		}
	}

	Path1Len = Path2Len = InLineIdx;	
	SET(Path1[Path1Len][0],Path1[Path1Len-1][1]);
	SET(Path1[Path1Len][1],Path1[Path1Len-1][0]);

	SET(Path2[Path2Len][0],Path2[Path2Len-1][1]);
	SET(Path2[Path2Len][1],Path2[Path2Len-1][0]);


	fprintf(LdrawFileOut, "0 Surface extruded\n");
	fprintf(LdrawFileOut, "0 Name: %s\n", argv[argc-1]);
	fprintf(LdrawFileOut, "0 Author: Philo's PathTruder\n0\n");
	fprintf(LdrawFileOut, "0 BFC CERTIFY CCW\n\n");
	
	printf("Normalize shape 1\n");

	len = DIST(Shape1[0][0],Shape1[0][1]);

	for(i=1; i<Shape1Len; i++)
	{
		SUB(Shape1[i][0], Shape1[i][0], Shape1[0][0]);
		MULT(Shape1[i][0], Shape1[i][0], (1/len));
		SUB(Shape1[i][1], Shape1[i][1], Shape1[0][0]);
		MULT(Shape1[i][1], Shape1[i][1], (1/len));
	}
	SUB(Shape1Vect, Shape1[0][1], Shape1[0][0]);

	Angle = atan2(-Shape1Vect[0], -Shape1Vect[1]);

	sa=sin(Angle); ca=cos(Angle);

	for(i=1; i<Shape1Len; i++)
	{
		Shape1[i-1][0][0] = Shape1[i][0][0] * ca - Shape1[i][0][1] * sa;
		Shape1[i-1][0][1] = Shape1[i][0][0] * sa + Shape1[i][0][1] * ca;
		Shape1[i-1][1][0] = Shape1[i][1][0] * ca - Shape1[i][1][1] * sa;
		Shape1[i-1][1][1] = Shape1[i][1][0] * sa + Shape1[i][1][1] * ca;
		Shape1[i-1][0][2] = Shape1[i][0][2];
		Shape1[i-1][1][2] = Shape1[i][1][2];
		if(invert)
		{
			Shape1[i-1][0][0] = -Shape1[i-1][0][0];
			Shape1[i-1][1][0] = -Shape1[i-1][1][0];
		}
	}
	Shape1Len--;

	printf("Normalize shape 2\n");

	len = DIST(Shape2[0][0],Shape2[0][1]);

	for(i=1; i<Shape2Len; i++)
	{
		SUB(Shape2[i][0], Shape2[i][0], Shape2[0][0]);
		MULT(Shape2[i][0], Shape2[i][0], (1/len));
		SUB(Shape2[i][1], Shape2[i][1], Shape2[0][0]);
		MULT(Shape2[i][1], Shape2[i][1], (1/len));
	}
	SUB(Shape2Vect, Shape2[0][1], Shape2[0][0]);

	Angle = atan2(-Shape2Vect[0], -Shape2Vect[1]);

	sa=sin(Angle); ca=cos(Angle);
	
	for(i=1; i<Shape2Len; i++)
	{
		Shape2[i-1][0][0] = Shape2[i][0][0] * ca - Shape2[i][0][1] * sa;
		Shape2[i-1][0][1] = Shape2[i][0][0] * sa + Shape2[i][0][1] * ca;
		Shape2[i-1][1][0] = Shape2[i][1][0] * ca - Shape2[i][1][1] * sa;
		Shape2[i-1][1][1] = Shape2[i][1][0] * sa + Shape2[i][1][1] * ca;
		Shape2[i-1][0][2] = Shape2[i][0][2];
		Shape2[i-1][1][2] = Shape2[i][1][2];
		if(invert)
		{
			Shape2[i-1][0][0] = -Shape2[i-1][0][0];
			Shape2[i-1][1][0] = -Shape2[i-1][1][0];
		}

	}
	Shape2Len--;

	
	printf("Extrusion\n");
	// Initialize current shape
	if(circular) endings=false;
	if(circular==1) n=Path1Len-1; else n=0;

	if(endings)
	{
		double Angle2=PathLocalBasis (0, 1, XVect, YVect, ZVect, Path1, Path2);
		Angle=PathLocalBasis (0, 0, XVect, YVect, ZVect, Path1, Path2);
		if(Angle2 > 90)
		{
			MULT(XVect, XVect, -1);
			MULT(ZVect, ZVect, -1);
		}
	}
	else
	{
		Angle=PathLocalBasis (n, 0, XVect, YVect, ZVect, Path1, Path2);
	}
	// compensate sharp angles
	if(compensate)
	{
		MULT(XVect, XVect, (1/cos(Angle*PI/360)));
	}

	// Calculate next transformed shape
	for(j=0; j<Shape1Len; j++)
	{
		for(k=0; k<2; k++)
		{
			MULT(NxtShape[j][k], XVect, Shape1[j][k][0]);
			MULT(temp1, YVect, Shape1[j][k][1]);
			ADD(NxtShape[j][k], NxtShape[j][k], temp1);	
			MULT(temp1, ZVect, Shape1[j][k][2]);
			ADD(NxtShape[j][k], NxtShape[j][k], temp1);	
			if(endings) 
			{
				ADD(NxtShape[j][k], NxtShape[j][k], Path1[1][0]);	
			}
			else 
			{
				ADD(NxtShape[j][k], NxtShape[j][k], Path1[0][0]);	
			}
		}

	}
	if(Angle > crease)
	{
		// sharp angle. Create line at junction
		for(i=0; i<Shape1Len; i++)
		{
			fprintf(LdrawFileOut, "2 24 %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", NxtShape[i][0][0], NxtShape[i][0][1], NxtShape[i][0][2], NxtShape[i][1][0], NxtShape[i][1][1], NxtShape[i][1][2]);
		}
	}

	int start, end;
	start=0; end=Path1Len;
	if(endings)
	{
		start++;
		end--;
	}
	for(i=start; i<end; i++)
	{

		// Transfer old next shape to current.
		for(j=0; j<Shape1Len; j++)
		{
			SET( CurShape[j][0],  NxtShape[j][0]);
			SET( CurShape[j][1],  NxtShape[j][1]);
		}
		
		if(i==end-1)
			if(circular==1)
				Angle = PathLocalBasis (i, 0, XVect, YVect, ZVect, Path1, Path2);
			else
			{
				if(endings)
				{
					double Angle2 = PathLocalBasis (i, i+1, XVect, YVect, ZVect, Path1, Path2);
					Angle = PathLocalBasis (i+2, i+2, XVect, YVect, ZVect, Path1, Path2);
					if(Angle2 < 90)
					{
						// in that case the local base is mirrorred...
						SUB (XVect, null, XVect);
						SUB (ZVect, null, ZVect);
					}
				}
				else
				{
					Angle = PathLocalBasis (i+1, i+1, XVect, YVect, ZVect, Path1, Path2);
					// in that case the local base is mirrorred...
					SUB (XVect, null, XVect);
					SUB (ZVect, null, ZVect);
				}
			}
		else
			Angle = PathLocalBasis (i, i+1, XVect, YVect, ZVect, Path1, Path2);

		// compensate sharp angles
		if(compensate)
		{
			MULT(XVect, XVect, (1/cos(Angle*PI/360)));
		}

		{
			double x;
			j=(i+1-start)*transitions % (2 * (end-start));
			x = 1.0 * j / (end-start);
			if (x>1) x=2-x;
//			printf("%d %lf\n", j, x);
			ratio = sigmoid(x, slope, position); 
		}

		
//		rotangle = rotation * PI / 180 * sigmoid((i+1.0) / Path1Len, rotslope, rotpos);
		rotangle = rotation * PI / 180 * (i+1.0) / Path1Len;
		
		sa = sin(rotangle); ca = cos(rotangle);

//		printf("%d %lf\n\n", i, ratio);

		for(j=0; j<Shape1Len; j++)
		{
			for(k=0; k<2; k++)
			{
				temp1[0] = Shape1[j][k][0] * ca - Shape1[j][k][1] * sa;
				temp1[1] = Shape1[j][k][0] * sa + Shape1[j][k][1] * ca;
				temp2[0] = Shape2[j][k][0] * ca - Shape2[j][k][1] * sa;
				temp2[1] = Shape2[j][k][0] * sa + Shape2[j][k][1] * ca;

				MULT(NxtShape[j][k], XVect, (temp1[0]*(1-ratio)+temp2[0]*ratio));
				MULT(temp3, YVect, (temp1[1]*(1-ratio)+temp2[1]*ratio));
				ADD(NxtShape[j][k], NxtShape[j][k], temp3);	
				MULT(temp3, ZVect, (Shape1[j][k][2]*(1-ratio)+Shape2[j][k][2]*ratio));
				ADD(NxtShape[j][k], NxtShape[j][k], temp3);	
				ADD(NxtShape[j][k], NxtShape[j][k], Path1[i+1][0]);	
			}
		}
		if(Angle > crease)
		{
			// sharp angle. Create line at junction
			for(j=0; j<Shape1Len; j++)
			{
				fprintf(LdrawFileOut, "2 24 %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", NxtShape[j][0][0], NxtShape[j][0][1], NxtShape[j][0][2], NxtShape[j][1][0], NxtShape[j][1][1], NxtShape[j][1][2]);
			}
		}
		// Generate tri/quad sheet
		for(j=0; j<Shape1Len; j++)
		{
			if ((DIST(Shape1[j][0], Shape1[j][1]) < EPSILON) && (DIST(Shape2[j][0], Shape2[j][1]) < EPSILON))
			{
				// Null lenth segment in shape file -> generate line at that place
				fprintf(LdrawFileOut, "2 24 %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", 
					CurShape[j][0][0], CurShape[j][0][1], CurShape[j][0][2],
					NxtShape[j][0][0], NxtShape[j][0][1], NxtShape[j][0][2]);
			}			
			if (DIST(CurShape[j][0], CurShape[j][1]) < VERTMERGE)
			{
				if ((DIST(NxtShape[j][0], NxtShape[j][1]) < VERTMERGE) || (DIST(CurShape[j][0], NxtShape[j][0]) < VERTMERGE) || (DIST(NxtShape[j][1], CurShape[j][1]) < VERTMERGE))
				{
					// Degenerated. Nothing to output
					continue;
				}
				else
				{
					fprintf(LdrawFileOut, "3 16 %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", 
						CurShape[j][0][0], CurShape[j][0][1], CurShape[j][0][2],
						NxtShape[j][1][0], NxtShape[j][1][1], NxtShape[j][1][2],
						NxtShape[j][0][0], NxtShape[j][0][1], NxtShape[j][0][2]);
					continue;
				}
			}
			if (DIST(NxtShape[j][0], NxtShape[j][1]) < VERTMERGE)
			{
				if ((DIST(CurShape[j][0], NxtShape[j][0]) < VERTMERGE) || (DIST(NxtShape[j][1], CurShape[j][1]) < VERTMERGE))
				{
					// Degenerated. Nothing to output
					continue;
				}
				else
				{
					fprintf(LdrawFileOut, "3 16 %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", 
						CurShape[j][0][0], CurShape[j][0][1], CurShape[j][0][2],
						CurShape[j][1][0], CurShape[j][1][1], CurShape[j][1][2],
						NxtShape[j][0][0], NxtShape[j][0][1], NxtShape[j][0][2]);
					continue;
				}
			}
			if (DIST(CurShape[j][0], NxtShape[j][0]) < VERTMERGE)
			{
				if (DIST(NxtShape[j][1], CurShape[j][1]) < VERTMERGE)
				{
					// Degenerated. Nothing to output
					continue;
				}
				else
				{
					fprintf(LdrawFileOut, "3 16 %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", 
						CurShape[j][0][0], CurShape[j][0][1], CurShape[j][0][2],
						CurShape[j][1][0], CurShape[j][1][1], CurShape[j][1][2],
						NxtShape[j][1][0], NxtShape[j][1][1], NxtShape[j][1][2]);
					continue;
				}
			}
			if (DIST(NxtShape[j][1], CurShape[j][1]) < VERTMERGE)
			{
				fprintf(LdrawFileOut, "3 16 %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", 
					CurShape[j][0][0], CurShape[j][0][1], CurShape[j][0][2],
					CurShape[j][1][0], CurShape[j][1][1], CurShape[j][1][2],
					NxtShape[j][0][0], NxtShape[j][0][1], NxtShape[j][0][2]);
				continue;
			}

			
			Angle=Tri_Angle(CurShape[j][0], NxtShape[j][0], NxtShape[j][1], CurShape[j][1] );
			if(Angle<SMALLANGLE)
			{
				fprintf(LdrawFileOut, "4 16 %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", 
					CurShape[j][0][0], CurShape[j][0][1], CurShape[j][0][2],
					CurShape[j][1][0], CurShape[j][1][1], CurShape[j][1][2],
					NxtShape[j][1][0], NxtShape[j][1][1], NxtShape[j][1][2],
					NxtShape[j][0][0], NxtShape[j][0][1], NxtShape[j][0][2]);
			}
			else
			{
				fprintf(LdrawFileOut, "3 16 %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", 
					CurShape[j][0][0], CurShape[j][0][1], CurShape[j][0][2],
					NxtShape[j][1][0], NxtShape[j][1][1], NxtShape[j][1][2],
					NxtShape[j][0][0], NxtShape[j][0][1], NxtShape[j][0][2]);
				fprintf(LdrawFileOut, "3 16 %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", 
					CurShape[j][0][0], CurShape[j][0][1], CurShape[j][0][2],
					CurShape[j][1][0], CurShape[j][1][1], CurShape[j][1][2],
					NxtShape[j][1][0], NxtShape[j][1][1], NxtShape[j][1][2]);
			}

		}
	}


	fclose (LdrawFileOut);
	printf("Press <Enter> to quit");
	getchar();
	return 0;

}